%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% File: warpDemo.m  
%%% Matlab script file 
%%%
%%% Fun with homographies:
%%% Compute a perspective image of a plane.  Use the observed
%%% horizon line to compute the affine reconstruction of the texture
%%% on the plane.  Finally, compute the similarity reconstruction
%%% using information about specific length ratios and angles.

%%% Dependencies
%%%   Directory: matlabVisTools
%%%      iseToolbox/pyrTools/
%%%           showIm.m 
%%%      utvisToolbox/file/
%%%          pgmRead.m resizeImageFig.m
%%%      

%%% CC and ADJ, Oct. '01.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear
close all


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Camera Parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% -----------------------------------------------
%% Intrinsic parameters

% # of pixels in a row = # of pixels in a col
nIm = 256;
szIm = [nIm nIm]; 
  
% optical axis piercing the image
o = (szIm + ones(1,2))/2 ; 

% aspect ratio set such that angle of view
% from the left side of the image to the center is 45 degrees
% ns/(2f) = tan(pi/4). assume pixels are square.
f = 1.65;  % The image is independent of f, because
           % we make the sizes of the pixels proportional to f...
s = 2*f /nIm;

% Intrinsic matrix
M_im = [f 0 0;
       0 f 0;
       0 0 1];

M_pix = [1/s   0     o(1);
	 0     1/s   o(2);
	 0     0      1];

M_int = M_pix*M_im;

%% -----------------------------------------------
%% Extrinsic parameters

% Put the camera along the world W_3-axis, looking back at the W_3=0 plane.

% define the rotation matrix R specifying 
% relative rotation between the two coordinate
% frames
thetaYZ = 0.2*pi;
ct = cos(thetaYZ);
st = sin(thetaYZ);
R1 = [1 0 0; 0 ct -st; 0 st ct];

thetaXZ = 3.5*pi/4;  % Almost looking backwards along W_3, or Z, axis
ct = cos(thetaXZ);
st = sin(thetaXZ);
R2 = [ct 0 -st; 0 1 0; st 0 ct];

thetaXY = 0;
ct = cos(thetaXY);
st = sin(thetaXY);
R3 = [ct -st 0; st ct 0; 0 0 1];
R = R1*R2*R3;

% origin of the world coordinate
% Put the camera far enough back to contain a 512x512 image centered
% on the world cooridinate frame origin.
T = [0 0 256]';
D = -R*T;

M_cam = [R D];



% -----------------------------------------------
% Fundamental matrix (homography)

H = M_int*M_cam

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Mapping coordinates to a world plane on the ground X_(w,3) = 0
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Original code in iseToolbox/utility/warpAffine2.m
% Code modified to illustrate homography

% Third coordinate X_(w,3) = 0, so delete the corresponding column of H.
F = H(:,[1,2,4]); 
A = pinv(F); % stable version of inverse

% size of the image
szIm = [nIm nIm];
% pixel coordinates
[x,y] = meshgrid(1:szIm(1),1:szIm(2));
pix   = [x(:)'; y(:)'];

% homogeneous pixels in warped frame.
hPixels = [ pix; ones(1,szIm(1)*szIm(2)) ];

% corresponding warped points in original frame
hScene  = A*hPixels;

% pixel coords in original frame.
xprime=(hScene(1,:)./(hScene(3,:)))';
yprime=(hScene(2,:)./(hScene(3,:)))';
xprime = reshape(xprime, szIm);
yprime = reshape(yprime, szIm);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Warping an image according to the homography
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Read input image
einIm = pgmRead('einstein.pgm');
% Place image at origin on W_3 = 0 plane.  We want
% it to look normal when we look down on it from W_3>0, so
% we need to flip it (the top right corner should go at the origin).
im = fliplr(einIm);

% Do the interplation, the xy value will be obtained from the
% interpolated value of im at the correpsonding (non-integer)
% location xprime,yprime.
result = interp2(x,y,im,xprime,yprime, '*linear');
skyGray = 0.8 * 256;
result(hScene(3,:) < 0) = skyGray;
result(isnan(result)) = 0;
result = reshape(result,szIm);

figure(1); clf; showIm(result);
pause;

%%  Finished perspective view of one image. %%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Tile the ground plane with these images...
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
xTile = mod(xprime, 256) + 1;
yTile = mod(yprime, 256) + 1;

result = interp2(x,y,im,xTile,yTile, '*linear');
skyGray = 0.8 * 256;
result(hScene(3,:) < 0) = skyGray;
result(isnan(result)) = 0;
result = reshape(result,szIm);

figure(1); clf; showIm(result);
pause;

%%%%%%%%%  Rotate the camera by increment the thetaYZ angle.
%%%%%%%%%  All other camera parameters are fixed.

for thetaYZ = 0.2*pi-0.2:-0.2:-0.3*pi

  ct = cos(thetaYZ);
  st = sin(thetaYZ);
  R1 = [1 0 0; 0 ct -st; 0 st ct];

  % The rest is the same as above...
  R = R1*R2*R3;
  T = [0 0 256]';
  D = -R*T;
  M_cam = [R D];
  H = M_int*M_cam;

  F = H(:,[1,2,4]);
  A = pinv(F); % stable version of inverse

  % pixel coordinates
  [x,y] = meshgrid(1:szIm(1),1:szIm(2));
  pix   = [x(:)'; y(:)'];

  % homogeneous pixels
  hPixels = [ pix; ones(1,szIm(1)*szIm(2)) ];
  % corresponding warped points
  hScene  = A*hPixels;

  xprime=(hScene(1,:)./(hScene(3,:)))';
  yprime=(hScene(2,:)./(hScene(3,:)))';

  xprime = reshape(xprime, szIm);
  yprime = reshape(yprime, szIm);

  % Tile the ground plane with these images...
  xTile = mod(xprime, 256) + 1;
  yTile = mod(yprime, 256) + 1;

  result = interp2(x,y,im,xTile,yTile, '*linear');
  skyGray = 0.8 * 256;
  result(hScene(3,:) < 0) = skyGray;
  result(isnan(result)) = 0;
  result = reshape(result,szIm);

  figure(1); clf; showIm(result);
  pause;

end  % End of loop over camera rotation
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Display the image to be used for the reconstruction
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
im = result;
nIm = size(im,1);
figure(1); clf; showIm(im);
resizeImageFig(1,size(im), 2);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Determine the horizon line
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
hrzn =    [ 3.0000   71.5000;  116.0000    3.5000];
%% Or mouse in two points on the horizon
%%hrzn = ginput(2);

%% Line constraint for horizon
t = (hrzn(1,:) - hrzn(2,:))';
t = t /sqrt( t' * t);
n = [t(2); -t(1)];
a =  -mean(hrzn * n);
% Swap signs so that a > 0. This makes our initial Hp matrix have determinant>0
if a < 0
  a = -a;
  n = -n;
end

%% Plot the horizon.
endPts = cropLineInBox(n, a, [0 0 size(im)]);
figure(1); hold on;
plot(endPts(:,1), endPts(:,2), 'r-', 'LineWidth', 2.0);
hold off;
pause;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Determine the homography for an affine reconstruction
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Hp = a * eye(3,3);
Hp(1:2, 3) = n;
Hp(3, 1:2) = -n';

%% Check that the horizon has been moved to the line at infinity
Hp' * [n; a]  % should be [ 0 0 *]'

%% Set the orientation of the transform, so it does not flip the image.

%% Bottom right corner of the data image is on the ground plane 
%% furthest from the horizon.
u0 = [256; 256; 1];

% Check for flipped reconstructions.
% A negative third component of (Hp^-1) * u0 causes a flip.
% A negative determinant of Hp also causes a flip.  Two flips
% annihilate each other.  
iHp = pinv(Hp);
q = iHp * u0;
if ( q(3) * det(Hp) < 0)  % There would be a flip
  Hp = Hp * diag([-1 1 1]); % Keep sign of q(3), flip sign of determinant
  iHp = pinv(Hp);
  q = iHp * u0;
end

%% Check that the horizon has still been moved to the line at infinity
Hp' * [n; a]  % should be [ 0 0 *]'


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Determine a similarity transform to scale and place the affine
%% reconstruction in the same sized image.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Choose the target location x0 for the corner point u0 to map into
%% in the affine reconstruction.
x0 = [50; 200; 1];

%% Build a scale and translation similarity transform in order
%% to relocate and scale the affine reconstruction:
Ha = eye(3,3);
Ha(1:2, 1:2) = Ha(1:2, 1:2) * 0.3;  
% smaller values for this scale factor (0.3) cause the affine reconstruction
% to expand.

%% Translation term d:
Ha(1:2, 3) = [0; 0];
%% Choose d in Ha such that u0 is proportional to Hp * Ha * x0
%% That is,  Hp inverse * u0  similar to Ha(with d=0)*x0 + [d; 0]
iHp = pinv(Hp);
q = iHp * u0;
r = Ha * x0;
d = q/q(3) - r/r(3);
%% Place translation d in affine transform matrix:
Ha(1:2, 3) = d(1:2);

%% Verify that x0 maps to u0
q = Hp * Ha * x0; q/q(3)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Do the image warp for affine reconstruction
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[x,y]=meshgrid(1:nIm,1:nIm);
coords=[x(:)'; y(:)' ; ones(1,length(x(:)))];
Qw = Hp * Ha * coords;
% Get image coords from the homogeneous coords
u = Qw(1,:)./Qw(3,:);
v = Qw(2,:)./Qw(3,:);

im2 = interp2(x,y,im, u, v, '*linear');
im2 = reshape(im2, nIm, nIm);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Display the affine reconstruction
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
im2(isnan(im2)) = 128;
figure(1); clf; showIm(im2);  %% Affine reconstruction 
resizeImageFig(1, size(im2), 3);
pause;

%% Notice the affine reconstruction has the tiles arranged along
%% parallel lines.  That is the effect of the homography Hp which
%% moved the horizon to the line at infinity.  
%% Also, notice the reconstructed tiles are noisy because of the
%% image sampling and interpolation.  
%% Finally, the reconstructed tiles are not in general square since
%% we have only done the reconstruction up to an affine transform of
%% the result.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Determine the remaining affine transformation for a similarity reconstruction
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Mouse in right angle, equal length axes
pts = [     36.1667   98.1667; 136.8333  183.8333; 246.8333  187.5000];
%% Suppose we know that in the correct reconstruction these two line segments are
%% of equal length and are at right angles to each other.
%%pts = ginput(3);

figure(1);hold on;
plot(pts(:,1), pts(:,2), 'b-o', 'LineWidth', 4.0);
hold off;
pause;

%% 
%% Choose A [1; 0] = (pts(3,:) - pts(2,:))/norm1;
%%        A [0; 1] = (pts(2,:) - pts(1,:))/norm2;
%% That is...
A = [(pts(3,:) - pts(2,:))' (pts(2,:) - pts(1,:))'];
A = A /(mean(diag(A' * A)).^0.5);

%% Build the corresponding affine transform matrix
Hb = [A [0 0]' ; 0 0 1];
%% Choose the translation so that the center point, pts(2,:), of
%% the coordinate axes is shifted to the center of the reconstructed image.
d =  [pts(2,:) 1]' - Hb * [nIm/2; nIm/2; 1];
Hb = [A d(1:2) ; 0 0 1];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Do the image warp for similarity reconstruction (use the original image, im)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[x,y]=meshgrid(1:nIm,1:nIm);
coords=[x(:)'; y(:)' ; ones(1,nIm*nIm)];
Qw =  Hp * Ha * Hb * coords;
%% Here (Hp * Ha) do the affine reconstruction of the original image, 
%% and Hb is the new transform to provide a similarity reconstruction.

% Get image coords from the homogeneous coords
u = Qw(1,:)./Qw(3,:);
v = Qw(2,:)./Qw(3,:);

% Do the interpolation
im3 = interp2(x,y,im, u, v, '*linear');
im3(isnan(im3)) = 128;
im3 = reshape(im3, nIm, nIm);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Display the similarity reconstruction
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
figure(1); clf; showIm(im3);  %% Similarity reconstruction 
hold on;
simPts = pInv(Hb) * [pts'; ones(1, size(pts,1))];
simPts = simPts ./ repmat(simPts(3,:),3,1);
simPts = simPts(1:2, :)';
hold on;
plot(simPts(:,1), simPts(:,2), 'b-o', 'LineWidth', 3.0);
hold off;
pause;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Summary: Display the initial image, along with the affine and
%% similarity reconstructions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
figure(1);clf;
subplot(2,2,1);
showIm(im); title('Original');
endPts = cropLineInBox(n, a, [0 0 size(im)]);
hold on;
plot(endPts(:,1), endPts(:,2), 'r-', 'LineWidth', 3.0);

perspPts = (Hp * Ha) * [pts'; ones(1, size(pts,1))];
perspPts = perspPts ./ repmat(perspPts(3,:),3,1);
perspPts = perspPts(1:2, :)';
plot(perspPts(:,1), perspPts(:,2), 'b-o', 'LineWidth', 3.0);
hold off;

subplot(2,2,2);
showIm(im2); title('Affine Reconstruction');
hold on;
plot(pts(:,1), pts(:,2), 'b-o', 'LineWidth', 3.0);
hold off;

subplot(2,2,4);
showIm(im3); title('Similarity Reconstruction');
hold on;
plot(simPts(:,1), simPts(:,2), 'b-o', 'LineWidth', 3.0);
hold off;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
